@@ -1,59 +1,72 @@
 import numpy as np
+import pyqmc.gpu as gpu
 import pyqmc.energy as energy
-from pyqmc.ewald import Ewald
+import pyqmc.ewald as ewald
+import pyqmc.eval_ecp as eval_ecp
+
+
+def gradient_generator(mol, wf, to_opt=None, nodal_cutoff=1e-3, **ewald_kwargs):
+    return PGradTransform(
+        EnergyAccumulator(mol, **ewald_kwargs), 
+        LinearTransform(wf.parameters, to_opt), 
+        nodal_cutoff=nodal_cutoff,
+    )
 
 
 class EnergyAccumulator:
-    """returns energy of each configuration in a dictionary. 
-  Keys and their meanings can be found in energy.energy """
+    """Returns local energy of each configuration in a dictionary."""
 
     def __init__(self, mol, threshold=10, **kwargs):
         self.mol = mol
         self.threshold = threshold
         if hasattr(mol, "a"):
-            self.ewald = Ewald(mol, **kwargs)
-
-            def compute_energy(mol, configs, wf, threshold):
-                ee, ei, ii = self.ewald.energy(configs)
-                ecp_val = energy.get_ecp(mol, configs, wf, threshold)
-                ke = energy.kinetic(configs, wf)
-                return {
-                    "ke": ke,
-                    "ee": ee,
-                    "ei": ei,
-                    "ecp": ecp_val,
-                    "total": ke + ee + ei + ecp_val + ii,
-                }
-
-            self.compute_energy = compute_energy
+            self.coulomb = ewald.Ewald(mol, **kwargs)
         else:
-            self.compute_energy = energy.energy
+            self.coulomb = energy.OpenCoulomb(mol, **kwargs)
 
     def __call__(self, configs, wf):
-        return self.compute_energy(self.mol, configs, wf, self.threshold)
+        ee, ei, ii = self.coulomb.energy(configs)
+        ecp_val = eval_ecp.ecp(self.mol, configs, wf, self.threshold)
+        ke, grad2 = energy.kinetic(configs, wf)
+        return {
+            "ke": ke,
+            "ee": ee,
+            "ei": ei,
+            "ecp": ecp_val,
+            "grad2": grad2,
+            "total": ke + ee + ei + ecp_val + ii,
+        }
 
     def avg(self, configs, wf):
-        d = {}
-        for k, it in self(configs, wf).items():
-            d[k] = np.mean(it, axis=0)
-        return d
+        return {k: np.mean(it, axis=0) for k, it in self(configs, wf).items()}
+
+    def nonlocal_tmoves(self, configs, wf, e, tau):
+        return eval_ecp.compute_tmoves(self.mol, configs, wf, e, self.threshold, tau)
+
+    def has_nonlocal_moves(self):
+        return self.mol._ecp != {}
 
     def keys(self):
-        return set(["ke", "ee", "ei", "ecp", "total"])
+        return set(["ke", "ee", "ei", "ecp", "total", "grad2"])
 
     def shapes(self):
-        return {"ke": (), "ee": (), "ei": (), "ecp": (), "total": ()}
+        return {"ke": (), "ee": (), "ei": (), "ecp": (), "total": (), "grad2": ()}
 
 
 class LinearTransform:
     """
-    Linearize a dictionary of wf parameters. 
-    to_opt is a dictionary with the keys to optimize, and its values are boolean arrays indicating which specific elements to optimize
-    Note: to_opt[k] can't be boolean scalar; it has to be an array with the same dimension as parameters[k]. 
-    to_opt doesn't have to have all the keys of parameters, but all keys of to_opt must be keys of parameters.
+    Linearize a dictionary of wf parameters.
+
+    :parameter dict parameters: the wave function parameters
+    :parameter dict to_opt: is a dictionary with the keys to optimize, and its values are boolean arrays indicating which specific elements to optimize
+
+    Note:
+        to_opt[k] can't be boolean scalar; it has to be an array with the same dimension as parameters[k].
+        to_opt doesn't have to have all the keys of parameters, but all keys of to_opt must be keys of parameters.
     """
 
     def __init__(self, parameters, to_opt=None):
+        parameters = {k: gpu.asnumpy(v) for k, v in parameters.items()}
         if to_opt is None:
             to_opt = {k: np.ones(p.shape, dtype=bool) for k, p in parameters.items()}
         self.to_opt = {k: o for k, o in to_opt.items() if np.any(o)}
@@ -71,16 +84,14 @@ class LinearTransform:
         self.nparams = np.sum([v.sum() for v in self.to_opt.values()])
 
     def serialize_parameters(self, parameters):
-        """Convert the dictionary to a linear list
-        of gradients
-        """
-        params = np.concatenate([parameters[k][opt] for k, opt in self.to_opt.items()])
+        """Convert the dictionary to a linear list of gradients"""
+        params = np.concatenate(
+            [gpu.asnumpy(parameters[k])[opt] for k, opt in self.to_opt.items()]
+        )
         return np.concatenate((params.real, params[self.complex_inds].imag))
 
     def serialize_gradients(self, pgrad):
-        """Convert the dictionary to a linear list
-        of gradients, mask allows for certain fixed parameters
-        """
+        """Convert the dictionary to a linear list of gradients, mask allows for certain fixed parameters"""
         grads = []
         for k, opt in self.to_opt.items():
             mask = ~np.repeat(opt[np.newaxis, :], pgrad[k].shape[0], axis=0)
@@ -91,8 +102,7 @@ class LinearTransform:
         return np.concatenate((grads, grads[:, self.complex_inds] * 1j), axis=1)
 
     def deserialize(self, parameters):
-        """Convert serialized parameters to dictionary
-        """
+        """Convert serialized parameters to dictionary"""
         n = 0
         m = self.nparams
         d = {}
@@ -107,7 +117,7 @@ class LinearTransform:
                 m_p = self.nimag[k]
                 flat_parms[opt_] += parameters[m : m + m_p] * 1j
                 m += m_p
-            d[k] = flat_parms.reshape(self.shapes[k])
+            d[k] = gpu.cp.asarray(flat_parms.reshape(self.shapes[k]))
             n += n_p
         return d
 
@@ -120,18 +130,14 @@ class PGradTransform:
         self.transform = transform
         self.nodal_cutoff = nodal_cutoff
 
-    def _node_regr(self, configs, wf):
-        """ 
-        Return true if a given configuration is within nodal_cutoff 
-        of the node 
-        Also return the regularization polynomial if true, 
-        f = a * r ** 2 + b * r ** 4 + c * r ** 3
+    def _node_regr(self, configs, grad2):
         """
-        ne = configs.configs.shape[1]
-        d2 = 0.0
-        for e in range(ne):
-            d2 += np.sum(np.abs(wf.gradient(e, configs.electron(e))) ** 2, axis=0)
-        r = 1.0 / d2
+        Return true if a given configuration is within nodal_cutoff
+        of the node
+        Also return the regularization polynomial if true,
+        f = a * r ** 2 + b * r ** 4 + c * r ** 6
+        """
+        r = 1.0 / grad2
         mask = r < self.nodal_cutoff ** 2
 
         c = 7.0 / (self.nodal_cutoff ** 6)
@@ -149,7 +155,7 @@ class PGradTransform:
         energy = d["total"]
         dp = self.transform.serialize_gradients(pgrad)
 
-        node_cut, f = self._node_regr(configs, wf)
+        node_cut, f = self._node_regr(configs, d["grad2"])
         dp_regularized = dp * f[:, np.newaxis]
 
         d["dpH"] = np.einsum("i,ij->ij", energy, dp_regularized)
@@ -173,7 +179,7 @@ class PGradTransform:
         energy = den["total"]
         dp = self.transform.serialize_gradients(pgrad)
 
-        node_cut, f = self._node_regr(configs, wf)
+        node_cut, f = self._node_regr(configs, den["grad2"])
 
         dp_regularized = dp * f[:, np.newaxis]
 
@@ -198,7 +204,7 @@ class PGradTransform:
 
 class SqAccumulator:
     r"""
-    Accumulates structure factor 
+    Accumulates structure factor
 
     .. math:: S(\vec{q}) = \langle \rho_{\vec{q}} \rho_{-\vec{q}} \rangle
                          = \langle \left| \sum_{j=1}^{N_e} e^{i\vec{q}\cdot\vec{r}_j} \right| \rangle
