@@ -1,4 +1,5 @@
 import numpy as np
+import pyqmc.gpu as gpu
 import scipy
 import h5py
 import os
@@ -31,11 +32,11 @@ def opt_hdf(hdf_file, data, attr, configs, parameters):
                 configs.initialize_hdf(hdf)
                 hdf.create_group("wf")
                 for k, it in parameters.items():
-                    hdf.create_dataset("wf/" + k, data=it)
+                    hdf.create_dataset("wf/" + k, data=gpu.asnumpy(it))
             hdftools.append_hdf(hdf, data)
             configs.to_hdf(hdf)
             for k, it in parameters.items():
-                hdf["wf/" + k][...] = it.copy()
+                hdf["wf/" + k][...] = gpu.asnumpy(it.copy())
 
 
 def polyfit_relative(xfit, yfit, degree):
@@ -46,48 +47,44 @@ def polyfit_relative(xfit, yfit, degree):
     return p, relative_error
 
 
-def stable_fit2(xfit, yfit, tolerance=1e-2):
-    """ Try to fit to a quadratic. If the fit is not good, 
-    then just take the lowest value of yfit
+def stable_fit(xfit, yfit, tolerance=1e-2):
+    """Fit a line and quadratic to xfit and yfit.
+    1. If the linear fit is as good as the quadriatic, choose the lower endpoint.
+    2. If the curvature is positive, estimate the minimum x value.
+    3. If the lowest yfit is less than the new guess, use that xfit instead.
+
+    :parameter list xfit: scalar step sizes along line
+    :parameter list yfit: estimated energies at xfit points
+    :parameter float tolerance: how good the quadratic fit needs to be
+    :returns: estimated x-value of minimum
+    :rtype: float
     """
     steprange = np.max(xfit)
     minstep = np.min(xfit)
+    a = np.argmin(yfit)
     pq, relative_errq = polyfit_relative(xfit, yfit, 2)
     pl, relative_errl = polyfit_relative(xfit, yfit, 1)
 
-    print("relative errors in fit", relative_errq, relative_errl)
     if relative_errl / relative_errq < 2:  # If a linear fit is about as good..
         if pl[0] < 0:
-            return steprange
+            est_min = steprange
         else:
-            return minstep
-    elif relative_errq < tolerance and pq[0] > 0:
+            est_min = minstep
+        out_y = np.polyval(pl, est_min)
+    elif relative_errq < tolerance and pq[0] > 0:  # If quadratic fit is good
         est_min = -pq[1] / (2 * pq[0])
         if est_min > steprange:
             est_min = steprange
         if est_min < minstep:
             est_min = minstep
-        return est_min
+        out_y = np.polyval(pq, est_min)
     else:
-        return xfit[np.argmin(yfit)]
-
-
-def stable_fit(xfit, yfit):
-    p = np.polyfit(xfit, yfit, 2)
-    steprange = np.max(xfit)
-    minstep = np.min(xfit)
-    est_min = -p[1] / (2 * p[0])
-    if est_min > steprange and p[0] > 0:  # minimum past the search radius
-        est_min = steprange
-    if est_min < minstep and p[0] > 0:  # mimimum behind the search radius
-        est_min = minstep
-    if p[0] < 0:
-        plin = np.polyfit(xfit, yfit, 1)
-        if plin[0] < 0:
-            est_min = steprange
-        if plin[0] > 0:
-            est_min = minstep
-    # print("estimated minimum adjusted", est_min, flush=True)
+        est_min = xfit[a]
+        out_y = yfit[a]
+    if (
+        out_y > yfit[a]
+    ):  # If min(yfit) has a lower energy than the guess, use it instead
+        est_min = xfit[a]
     return est_min
 
 
@@ -96,8 +93,8 @@ def line_minimization(
     coords,
     pgrad_acc,
     steprange=0.2,
-    warmup=0,
     max_iterations=30,
+    warmup_options=None,
     vmcoptions=None,
     lmoptions=None,
     update=sr_update,
@@ -111,35 +108,19 @@ def line_minimization(
     """Optimizes energy by determining gradients with stochastic reconfiguration
         and minimizing the energy along gradient directions using correlated sampling.
 
-    Args:
-
-      :wf: initial wave function
-
-      :coords: initial configurations
-
-      :pgrad_acc: A PGradAccumulator-like object
-
-      :steprange: How far to search in the line minimization
-
-      :warmup: number of steps to use for vmc warmup
-
-      :max_iterations: (maximum) number of steps in the gradient descent
-
-      :vmcoptions: a dictionary of options for the vmc method
-
-      :lmoptions: a dictionary of options for the lm method
-
-      :update: A function that generates a parameter change 
-
-      :update_kws: Any keywords 
-
-      :npts: number of points to fit to in each line minimization
-
-    Returns:
-
-      :wf: optimized wave function
-
-
+    :parameter wf: initial wave function
+    :parameter coords: initial configurations
+    :parameter pgrad_acc: A PGradAccumulator-like object
+    :parameter float steprange: How far to search in the line minimization
+    :parameter int warmup: number of steps to use for vmc warmup
+    :parameter int max_iterations: (maximum) number of steps in the gradient descent
+    :parameter dict vmcoptions: a dictionary of options for the vmc method
+    :parameter dict lmoptions: a dictionary of options for the lm method
+    :parameter update: A function that generates a parameter change
+    :parameter update_kws: Any keywords
+    :parameter int npts: number of points to fit to in each line minimization
+    :parameter boolean verbose: print output if True
+    :return: optimized wave function
     """
 
     if vmcoptions is None:
@@ -149,6 +130,11 @@ def line_minimization(
         lmoptions = {}
     if update_kws is None:
         update_kws = {}
+    if warmup_options is None:
+        warmup_options = dict(nblocks=1, nsteps_per_block=10, verbose=verbose)
+    if "tstep" not in warmup_options and "tstep" in vmcoptions:
+        warmup_options["tstep"] = vmcoptions["tstep"]
+    assert npts >= 3, f"linemin npts={npts}; need npts >= 3 for correlated sampling"
 
     # Restart
     iteration_offset = 0
@@ -157,7 +143,7 @@ def line_minimization(
             if "wf" in hdf.keys():
                 grp = hdf["wf"]
                 for k in grp.keys():
-                    wf.parameters[k] = np.array(grp[k])
+                    wf.parameters[k] = gpu.cp.asarray(grp[k])
             if "iteration" in hdf.keys():
                 iteration_offset = np.max(hdf["iteration"][...]) + 1
 
@@ -174,7 +160,7 @@ def line_minimization(
             accumulators={"pgrad": pgrad_acc},
             client=client,
             npartitions=npartitions,
-            **vmcoptions
+            **vmcoptions,
         )
         en = np.real(np.mean(df["pgradtotal"], axis=0))
         en_err = np.std(df["pgradtotal"], axis=0) / np.sqrt(df["pgradtotal"].shape[0])
@@ -203,8 +189,10 @@ def line_minimization(
         accumulators={},
         client=client,
         npartitions=npartitions,
-        **vmcoptions
+        **warmup_options,
     )
+    if verbose:
+        print("finished warmup", flush=True)
     df = []
     # Gradient descent cycles
     for it in range(max_iterations):
@@ -229,7 +217,7 @@ def line_minimization(
         # include near zero in the fit, and go backwards as well
         # We don't use the above computed value because we are
         # doing correlated sampling.
-        steps = np.linspace(-steprange / npts, steprange, npts)
+        steps = np.linspace(-steprange / (npts - 2), steprange, npts)
         params = [x0 + update(pgrad, Sij, step, **update_kws) for step in steps]
         if client is None:
             stepsdata = correlated_compute(wf, coords, params, pgrad_acc)
@@ -261,24 +249,21 @@ def line_minimization(
 
 
 def correlated_compute(wf, configs, params, pgrad_acc):
-    """ 
+    """
     Evaluates accumulator on the same set of configs for correlated sampling of different wave function parameters
 
-    Args:
-        :wf: wave function object
-        :configs: (nconf, nelec, 3) array
-        :params: (nsteps, nparams) array 
-            list of arrays of parameters (serialized) at each step
-
-        :pgrad_acc: PGradAccumulator 
-
-    Returns:
-        :data: a single dict with indices [parameter, values]
+    :parameter wf: wave function object
+    :parameter configs: (nconf, nelec, 3) array
+    :parameter params: (nsteps, nparams) array
+        list of arrays of parameters (serialized) at each step
+    :parameter pgrad_acc: PGradAccumulator
+    :returns: a single dict with indices [parameter, values]
 
     """
 
     data = []
     psi0 = wf.recompute(configs)[1]  # recompute gives logdet
+
     for p in params:
         newparms = pgrad_acc.transform.deserialize(p)
         for k in newparms:
