@@ -47,10 +47,16 @@ class Parameters:
             for k2 in self.data[k1].keys():
                 yield k1 + k2
 
+    def values(self):
+        for i in range(self.wf_count):
+            k1 = "wf" + str(i + 1)
+            for k2 in self.data[k1].keys():
+                yield self.data[k1][k2]
+
 
 class MultiplyWF:
     """
-    A general representation of a wavefunction as a product of multiple wf_factors 
+    A general representation of a wavefunction as a product of multiple wf_factors
     """
 
     def __init__(self, *wf_factors):
@@ -89,6 +95,11 @@ class MultiplyWF:
         testvalues = [wf.testvalue_many(e, epos, mask=mask) for wf in self.wf_factors]
         return np.prod(testvalues, axis=0)
 
+    def gradient_value(self, e, epos):
+        grad_vals = [wf.gradient_value(e, epos) for wf in self.wf_factors]
+        grads, vals = list(zip(*grad_vals))
+        return np.sum(grads, axis=0), np.prod(vals, axis=0)
+
     def gradient_laplacian(self, e, epos):
         grad_laps = [wf.gradient_laplacian(e, epos) for wf in self.wf_factors]
         grads, laps = list(zip(*grad_laps))
@@ -107,8 +118,6 @@ class MultiplyWF:
 
 
 def test_parameters():
-    import numpy as np
-
     dicts = [{"coeff" + str(i): np.random.rand(3)} for i in range(10)]
     p = Parameters(dicts)
     # test len
